from functools import wraps
from starlette.datastructures import Headers
from Osdental.Models.Response import Response
from Osdental.Models.FakeInfo import FakeInfo
from Osdental.Shared.Enums.GrahpqlOperation import GraphqlOperation

# Client
def with_grpc_metadata(func):
    """
    Decorator to transform data to bytes and build metadata for gRPC.
    """
    @wraps(func)
    async def wrapper(self, request, headers: Headers, *args, **kwargs):
        # Build metadata from headers
        user_token = headers.get('authorization', '')
        metadata = [
            ('authorization', user_token),
            ('dynamicclientid', headers.get('dynamicClientId') or None),
        ]

        # Add extra metadata if it comes in kwargs
        extra_metadata = kwargs.pop('extra_metadata', None)
        if extra_metadata:
            if isinstance(extra_metadata, dict):
                metadata.extend(extra_metadata.items())
            else:
                metadata.extend(extra_metadata)

        # Call the actual gRPC method passing request and metadata
        res = await func(self, request, metadata, *args, **kwargs)

        # Return a uniform Response
        return Response(status=res.status, message=res.message, data=res.data)

    return wrapper

# Server
def with_grpc_info(operation_name: GraphqlOperation = GraphqlOperation.QUERY):
    """
    Decorator to wrap gRPC methods and build the FakeInfo.
    """
    def decorator(func):
        @wraps(func)
        async def wrapper(servicer, request, context):
            metadata_dict = dict(context.invocation_metadata())
            user_token = metadata_dict.get('authorization')
            token_value = user_token.split(' ')[1] if user_token.startswith('Bearer ') else user_token
            context_dict = {
                'user_token': token_value,
                'headers': metadata_dict,
            }
            info = FakeInfo(context_dict, operation_name)

            return await func(servicer, request.data, info)
        return wrapper
    return decorator


import asyncio
from typing import Dict, Any
from contextlib import asynccontextmanager
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.exc import OperationalError, IntegrityError, DBAPIError, InvalidRequestError
from Osdental.Exception.ControlledException import OSDException, DatabaseException
from Osdental.Shared.Enums.Code import Code

class UnitOfWork:
    def __init__(self, session: AsyncSession, repositories: Dict[str, Any] = None):
        self.session = session
        self._repositories = repositories or {}
        self._instances = {}

    def __getattr__(self, item):
        if item in self._instances:
            return self._instances[item]
        
        raise OSDException(error=f'UnitOfWork object has no attribute {item}')


    @asynccontextmanager
    async def __call__(self):
        try:
            async with self.session.begin():
                self._instances = {name: repo_cls(self.session) for name, repo_cls in self._repositories.items()}
                try:
                    yield self
                except Exception:
                    await self.session.rollback()
                    raise
                else:
                    await self.session.commit()

        except asyncio.CancelledError:
            raise 
        except IntegrityError as ierr:
            raise DatabaseException(status_code=Code.INTEGRITY_DB_ERROR_CODE, error=str(ierr))
        except OperationalError as oerr:
            raise DatabaseException(status_code=Code.OPERATIONAL_DB_ERROR_CODE, error=str(oerr))
        except (DBAPIError, InvalidRequestError) as db_err:
            raise DatabaseException(status_code=Code.DB_UNKNOWN_ERROR_CODE, error=str(db_err))
        except DatabaseException:
            raise
        except Exception as err:
            raise OSDException(error=str(err))
        finally:
            self._instances.clear()
